final int BULLET = 0;
final int MISSILE = 1;
final int BOUND = 2;
final int THREEWAY = 3;
final int FOURWAY = 4;
final int SCATTER = 5;
final int PLANE = 6;
final int NIGHTWING = 7;
final int ICEBERG = 8;
final int SWINE = 9;
abstract class Enemy {
  Shape shape;
  Enemy parent;
  Moving moving;
  float fx ,fy ,fz;
  float x ,y ,z;
  float vx ,vy;
  float sx ,sy ,sz;
  float largeness;
  float power;
  float maxhp;
  float hp;
  int attackTerm;
  int reloadTerm;
  int vanishTerm;
  boolean start;
  private int attackCnt;
  private int reloadCnt;
  private int attackTimes;
  private float dam;
  private boolean exist;
  private boolean destroyed;
  private boolean attacking;
  private boolean vanished;
  private int vanCnt;
  int cnt;
  Enemy(Enemy parent ,Moving moving,float hp ,float fx ,float fy ,float fz ,float vx ,float vy){
    this.parent = parent;
    this.moving = moving;
    this.fx = fx;
    this.fy = fy;
    this.fz = fz;
    this.vx = vx;
    this.vy = vy;
    this.maxhp = hp;
    this.hp = maxhp;
    if(maxhp < 5)largeness = 15;
    else largeness = sqrt(maxhp - 4) * 5 + 15;
    x = field.transX(fx ,fy ,fz);
    y = field.transY(fx ,fy ,fz);
    z = field.transZ(fx ,fy ,fz);
    exist =true;
    destroyed = false;
    vanished = false;
    start = true;
    attacking = true;
    
    attackCnt = 0;
    reloadCnt = 0;
    attackTimes = 0;
    dam = 0;
    cnt = 0;
    vanCnt = 0;
  }
  void update(){
    if(0 < vanishTerm && vanishTerm <= cnt)vanish();
   if(vanished){
     if(30 < vanCnt)setExist(false);
      vanCnt ++; 
   }else if(start){
     float fr = dist(fx ,fy ,field.x ,field.y );
     if(field.radius < fr){
       fx = fx * 0.99;
       fy = fy * 0.99;
     }else if(5 < abs(fz)){
       fz -= fz /abs(fz) * 5;
     }else {
       fz = 0;
       start = false;
     }
   }else {
     if(attacking){
       
       attackCnt ++;
       attack(attackTimes ,attackCnt);
       if(attackTerm < attackCnt){
        attacking = false;
        attackCnt = 0;
        attackTimes ++;
       }
       
      } else{
      reloadCnt ++;
      if(reloadTerm < reloadCnt){
        if(0 < attackTerm)attacking = true;
        reloadCnt = 0;
       } 
     }
     moving.update(attacking);
     float fr = dist(fx ,fy ,field.x ,field.y );
     if(field.radius < fr){
       bonk(fr);
     }
     if(dist(ship.x ,ship.y ,ship.z ,x ,y ,z) < largeness){
        ship.damage(power);
     }
   }
    x = field.transX(fx ,fy ,fz);
    y  = field.transY(fx ,fy ,fz);
    z = field.transZ(fx ,fy ,fz);
    float[] ax = shape.x;
    float[] ay = shape.y;
    float[] az = shape.z;
    sx = dam * random(5.0f);
    sy = dam * random(5.0f);
    sz = dam * random(5.0f);
    noFill();
    translate(x + sx, y + sy, z + sz);
    roll();
    if(start)stroke(122);
    else if(vanished){
      if(destroyed){
        stroke(122);
        scale(1 + (900.0 - (30 - (float)vanCnt) * (30 - (float)vanCnt)) / 100.0);
      }else{
        stroke(122);
        scale(1.0 - (900.0 - (30 - (float)vanCnt) * (30 - (float)vanCnt)) / 900.0);
      }
    }
    else stroke(255);
    if(parent == null)strokeWeight(2);
    else strokeWeight(1);
    beginShape(LINE_LOOP);
    for(int i = 0; i < ax.length ;i ++){
      vertex(ax[i] ,ay[i] ,az[i]);
    }
    endShape();
    strokeWeight(1);
    if(dam > 0)dam -= 1.0f;
    else dam = 0;
    cnt ++;
  }
   void setExist(boolean e){
    exist = e;
  }
  void destroy(){
    float rec;
    if(parent == null){
      stage.wait -= (float)maxhp * stage.rank / 2;
      rec = log(maxhp + 1);
    }else rec = log(maxhp + 1) / 10;
    ship.recPlus(rec);
    destroyed = true;
    vanished = true;
  }
  void vanish(){
    destroyed = false;
    vanished = true;
  }
   boolean isExist(){
    return exist;
  }
  float damage(float power){
    float h;
    if(start || destroyed)h = 0;
    else{
      h =  hp;
      hp -= power;
      dam = power;
      if(hp <= 0)destroy();
    }
    return h;
  }
  float aim(){
    return atan2(vy ,vx);
  }
  abstract void attack(int attackTerm ,int attackCnt);
  abstract void bonk(float fr);
  abstract void roll();
}
Shape BULLET_SHAPE;
class Bullet extends Enemy{
  Bullet(Enemy parent ,Moving moving ,float hp ,float fx ,float fy ,float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    reloadTerm = 0;
    attackTerm = 0;
    vanishTerm = -1;
    power = 5.0;
    if(BULLET_SHAPE == null){
      float[] a = {-2.2 ,-0.8 ,0 ,0.8 ,1.0};
      float[] b = {0  ,0.6 ,1 ,0.6 ,0};
      BULLET_SHAPE = new Pole(a ,b ,6);
    }
    shape = (Shape)BULLET_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){}
  void bonk(float fr){
    vanish();
  }
  void roll(){
    rotateZ(aim());
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
  }
}
Shape MISSILE_SHAPE;
class Missile extends Enemy {
  Missile(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 0;
    attackTerm = 0;
    vanishTerm = 300;
    if(MISSILE_SHAPE == null){
      float[] a = {-1 ,-1.5  ,-1   ,-0.5 ,0 ,2   ,2.5};
      float[] b = {0  ,1     ,0.7  ,0.3  ,1 ,0.7 ,0};
      MISSILE_SHAPE = new Pole(a ,b ,6);
    }
    shape = (Shape)MISSILE_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){}
  void bonk(float fr){
  }
  void roll(){
    rotateZ(aim());
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
  }
}
Shape BOUND_SHAPE;
class Bound extends Enemy {
  Bound(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 5.0;
    reloadTerm = 0;
    attackTerm = 0;
    vanishTerm = 600;
    if(BOUND_SHAPE == null){
      float[] a = {-1 ,0 ,1};
      float[] b = {0 ,1 ,0};
      BOUND_SHAPE = new Pole(a ,b ,6);
    }
    shape = (Shape)BOUND_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){}
  void bonk(float fr){
    float vr = dist(0 ,0 ,vx ,vy);
    float theta = atan2(fy ,fx);
    float rad = atan2(vy ,vx);
    vx = - vr * cos(2 * theta -rad);
    vy = - vr * sin(2 * theta - rad);
    
    fx = fx / fr * field.radius;
    fy = fy / fr * field.radius;
  }
  void roll(){
    rotateY(HALF_PI);
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
  }
}
Shape THREEWAY_SHAPE;
class Threeway extends Enemy {
  Threeway(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 60;
    attackTerm = 30;
    vanishTerm = 600;
    if(THREEWAY_SHAPE == null){
      float[] a = {-1.2 ,-1  ,-0.6   ,0  ,0.6 ,1   ,1.2};
      float[] b = {0  ,0.5     ,0.8  ,1 ,0.8  ,0.5 ,0};
      THREEWAY_SHAPE = new Pole(a ,b ,6);
    }
    shape = (Shape)THREEWAY_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){
    if(attackCnt % 30 == 1){
      float aim = atan2(ship.fy - fy ,ship.fx -fx);
      ep.addEnemy(this ,BULLET ,new Straight() ,30 ,fx ,fy ,fz ,cos(aim) * 5.0,sin(aim) * 5.0);
      ep.addEnemy(this ,MISSILE ,new Homing(TWO_PI / 360) ,10 ,fx ,fy ,fz ,cos(aim + PI / 3) * 3.0 ,sin(aim + PI / 3) * 3.0);
      ep.addEnemy(this ,MISSILE ,new Homing(TWO_PI / 360) ,10 ,fx ,fy ,fz ,cos(aim - PI / 3) * 3.0 ,sin(aim - PI / 3) * 3.0);
    }
  }
  void bonk(float fr){
  }
  void roll(){
    rotateY(HALF_PI);
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
  }
}
Shape FOURWAY_SHAPE;
class Fourway extends Enemy {
  Fourway(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 60;
    attackTerm = 30;
    vanishTerm = 600;
    if(FOURWAY_SHAPE == null){
      float[][] a = {
        {-1.2 ,-0.6 ,0   ,0.6 ,1.2},
        {-0.4 ,-0.4 ,-0.4  ,0.4 ,0.4}
      };
      float[][] b ={
        {0    ,0.6  ,1.2 ,0.6 ,0},
        {0.4  ,0.4  ,0.4   ,0.4  ,0.4}
      };
      FOURWAY_SHAPE = new Pole(a ,b ,8);
    }
    shape = (Shape)FOURWAY_SHAPE.clone();
    shape.multi(largeness * 1.5);
  }
  void attack(int attackTerm ,int attackCnt){
    if(attackCnt % 30 == 1){
      float aim = TWO_PI * (float)(cnt % 180) / 180;
      ep.addEnemy(this ,BOUND ,new Straight() ,10 ,fx ,fy ,fz ,cos(aim) * 5.0,sin(aim) * 5.0);
      ep.addEnemy(this ,BOUND ,new Straight() ,10 ,fx ,fy ,fz ,cos(aim + PI / 2) * 5.0 ,sin(aim + PI / 3) * 5.0);
      ep.addEnemy(this ,BOUND ,new Straight() ,10 ,fx ,fy ,fz ,cos(aim + PI) * 5.0 ,sin(aim + PI / 3) * 5.0);
      ep.addEnemy(this ,BOUND ,new Straight() ,10 ,fx ,fy ,fz ,cos(aim - PI / 2) * 5.0 ,sin(aim - PI / 3) * 5.0);
    }
  }
  void bonk(float fr){
  }
  void roll(){
    rotateY(HALF_PI);
    rotateX(TWO_PI * (float)(cnt % 180) / 180);
  }
}
Shape SCATTER_SHAPE;
class Scatter extends Enemy {
  Scatter(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 60;
    attackTerm = 30;
    vanishTerm = 600;
    if(SCATTER_SHAPE == null){
      float[][] a = {
        {-0.6 ,-1 ,-0.6  ,1.0 ,0.4},
        {-0.6 ,0  ,0     ,0  ,0.4}
      };
      float[][] b ={
       {0 ,0.5  ,1.0 ,0.4 ,0},
       {0 ,0.6  ,0.6  ,0.6,0}
     };
      SCATTER_SHAPE = new Pole(a ,b ,6);
    }
    shape = (Shape)SCATTER_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){
    if(attackCnt % 10 == 0){
      float drad = PI / 30.0;
      float aim = atan2(vy ,vx);
      float si ,co;
      float dsi ,dco;
      float temp;
      si = sin(aim - PI / 3.0);
      co = cos(aim - PI / 3.0);
      dsi = sin(drad);
      dco = cos(drad);
      for(float i = -PI /3.0;i <= PI / 3.0;i += drad){
        temp = co;
        co = co * dco - si * dsi;
        si = temp * dsi + si * dco;
        ep.addEnemy(this ,BULLET ,new Straight() ,1 ,fx ,fy ,fz ,co * 3.0,si * 3.0);
      }
    }
  }
  void bonk(float fr){
  }
  void roll(){
    
    rotateZ(aim());
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
  }
}
Shape PLANE_SHAPE;
class Plane extends Enemy {
  Plane(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 60;
    attackTerm = 30;
    vanishTerm = 600;
    if(PLANE_SHAPE == null){
      float[][] a = {
        {-1.0 ,-0.8 ,-0.6 ,0.2 ,1.4},
        {-1.0 ,-1.0 ,-0.6 ,0.2 ,0.2}
      };
      float[][] b ={
       {0 ,0.8  ,1.2 ,0.6 ,0.0},
       {0.2 ,0.2  ,0.4 ,0.2 ,0.2}
     };
      PLANE_SHAPE = new Pole(a ,b ,4);
    }
    shape = (Shape)PLANE_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){
    if(attackCnt % 5 == 0 && 15 < attackCnt ){
      float vr = dist(0 ,0 ,vx ,vy);
      ep.addEnemy(this ,BULLET ,new Straight() ,5 ,fx ,fy ,fz ,vx / vr * 6.0 ,vy / vr *6.0);
    }
  }
  void bonk(float fr){
  }
  void roll(){
    rotateZ(aim());
  }
}
Shape NIGHTWING_SHAPE;
class Nightwing extends Enemy {
  Nightwing(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 60;
    attackTerm = 60;
    vanishTerm = 600;
    if(NIGHTWING_SHAPE == null){
      float[][] a = {
        {0 ,0.2 ,0   ,-1.0 ,0.8 ,1.0 ,0.9 ,0.7},
        {0 ,0.2 ,0.2 ,0.2  ,0.5 ,0.5 ,0.5 ,0.0}
      };
      float[][] b ={
       {0  ,0.4 ,0.8 ,1.2  ,1.0 ,0.6 ,0.2 ,0.0},
       {0  ,0.3  ,0.5 ,0.6 ,0.5 ,0.5 ,0.3 ,0.0}
     };
      NIGHTWING_SHAPE = new Pole(a ,b ,8);
    }
    shape = (Shape)NIGHTWING_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){
    if(20 < attackCnt){
      if(attackCnt < 40){
        if(attackCnt % 3 == 0){
          float vr = dist(0 ,0 ,vx ,vy);
          ep.addEnemy(this ,BULLET ,new Straight() ,5 ,fx ,fy ,fz ,vx / vr * 8.0 ,vy / vr * 8.0);
        }
      }else if(attackCnt == 55){
        float vr = dist(0 ,0 ,vx ,vy);
        ep.addEnemy(this ,MISSILE ,new Homing(TWO_PI / 180.0) ,20 ,fx ,fy ,fz ,- vx / vr * 5.0 ,- vy / vr * 5.0);
      }
    }
  }
  void bonk(float fr){
  }
  void roll(){
    rotateZ(aim());
    rotateY(- HALF_PI * max((1 - float(max(cnt - 80,0)) / 30.0) ,0));
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
    
  }
}
Shape ICEBERG_SHAPE;
class Iceberg extends Enemy {
  Iceberg(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 60;
    attackTerm = 30;
    vanishTerm = 600;
    if(ICEBERG_SHAPE == null){
      float[][] a = {
        {0 ,-0.5 ,0 ,1.0 ,0},
        {0}
      };
      float[][] b ={
       {0  ,0.5 ,1.0 ,0.5 ,0.0},
       {0.5}
     };
      ICEBERG_SHAPE = new Pole(a ,b ,8);
    }
    shape = (Shape)ICEBERG_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){
    if(attackCnt % 10 == 0){
      float fr = dist(fx ,fy ,ship.fx ,ship.fy);
      ep.addEnemy(this ,BOUND ,new Straight() ,5 ,fx ,fy ,fz , (ship.fx -fx) / fr * 5.0 ,(ship.fy -fy) / fr * 5.0);
    }
  }
  void bonk(float fr){
  }
  void roll(){
    rotateY(- HALF_PI);
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
    
  }
}
Shape SWINE_SHAPE;
class Swine extends Enemy {
  Swine(Enemy parent ,Moving moving ,float hp ,float fx ,float fy, float fz ,float vx ,float vy){
    super(parent ,moving ,hp ,fx ,fy ,fz ,vx ,vy);
    power = 10.0;
    reloadTerm = 45;
    attackTerm = 20;
    vanishTerm = 600;
    if(SWINE_SHAPE == null){
      float[][] a = {
        {-1.0 ,-0.8 ,-0.6 ,0.2 ,1.2 ,0.2},
        {-1.0 ,-0.8 ,-0.6 ,0.2 ,0.4 ,0.4}
      };
      float[][] b ={
       {0 ,0.8  ,1.2 ,0.6 ,0.3 ,0},
       {0 ,0.4  ,0.6 ,0.4 ,0.4 ,0.2}
     };
      SWINE_SHAPE = new Pole(a ,b ,4);
    }
    shape = (Shape)SWINE_SHAPE.clone();
    shape.multi(largeness);
  }
  void attack(int attackTerm ,int attackCnt){
  }
  void bonk(float fr){
  }
  void roll(){
     rotateZ(aim());
    rotateX(TWO_PI * (float)(cnt % 60) / 60);
    
  }
}
class EnemyPool {
  private ArrayList enemies;
  EnemyPool(){
    init();
  }
  void init(){
    if(enemies != null){
      enemies.clear();
    }else enemies = new ArrayList();
  }
  void addEnemy(Enemy parent ,int type ,Moving moving ,float hp ,float x ,float y ,float z ,float vx ,float vy){
    Enemy e;
    switch(type){
      case BULLET:   e =new Bullet(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case MISSILE:  e =new Missile(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case BOUND:    e =new Bound(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case THREEWAY: e =new Threeway(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case FOURWAY:  e =new Fourway(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case SCATTER:  e =new Scatter(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case PLANE:    e =new Plane(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case NIGHTWING:e =new Nightwing(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case ICEBERG:  e =new Iceberg(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      case SWINE:    e =new Swine(parent ,moving ,hp ,x ,y ,z ,vx ,vy);break;
      default:
      e =null;break;
    }
    moving.set(e);
    enemies.add(e);
  }
  void update(){
    int i = enemies.size() - 1;
      while(0 <=i){
        Enemy e = (Enemy)enemies.get(i);
        popMatrix();
        pushMatrix();
        e.update();
        if(e.isExist() == false){
          enemies.remove(i);
        }
        i--;
      }
    }
    void allDamage(float power){
      int i = enemies.size() - 1;
      while(0 <=i){
        Enemy e = (Enemy)enemies.get(i);
        e.damage(power);
        i--;
      }
    }
}
